# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pickle
import random
import string
from time import sleep, time
from typing import Tuple

import numpy as np
from linien_common.common import PSDAlgorithm
from linien_server.optimization.engine import MultiDimensionalOptimizationEngine
from pylpsd import lpsd
from scipy import signal

ALL_DECIMATIONS = list(range(32))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def calculate_psd(
    sig: np.ndarray, fs: float, algorithm: PSDAlgorithm
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the power spectral density.

    :param sig: The signal to calculate the PSD for.
    :param fs: The sampling frequency.
    :param algorithm: The PSD algorithm to use.
    :return: One-sided power spectral density.
    """

    # at beginning or end of signal, we sometimes have more glitches --> ignore
    # them (200 points less @ 16384 points doesn't hurt much)
    sig = sig[100:-100]

    num_pts = 256
    window = "hann"  # passed to scipy.signal.get_window for welch and lpsd

    sig = sig.astype(np.float64)

    if algorithm == PSDAlgorithm.WELCH:
        f, Pxx = signal.welch(
            sig, fs, window=window, nperseg=num_pts, scaling="density"
        )
    elif algorithm == PSDAlgorithm.LPSD:
        fmin = fs / len(sig) * 10  # lowest frequency of interest
        fmax = fs / 20.0  # highest frequency of interest

        f, Pxx = lpsd(
            sig,
            fs,
            window=window,
            fmin=fmin,
            fmax=fmax,
            Jdes=256,
            Kmin=2,
            scaling="density",
        )
    return f, Pxx


def residual_freq_noise(dt, sig, algorithm):
    fs = 1 / dt

    f, psd = calculate_psd(sig, fs, algorithm)
    # we want to have it in counts / Sqrt[Hz], not in (counts**2) / Hz
    psd = np.sqrt(psd)

    return f, psd


def psds_to_fitness(psds_by_decimation):
    fitness = 0

    for decimation, psd_data in psds_by_decimation.items():
        f, psd = psd_data
        fitness += sum(psd)

    return fitness


def generate_curve_uuid():
    return "".join(random.choice(string.ascii_lowercase) for i in range(10))


class PSDAcquisition:
    def __init__(self, control, parameters, is_child=False):
        self.decimation_index = 0

        self.recorded_signals_by_decimation = {}
        self.recorded_psds_by_decimation = {}

        self.running = True
        self.parameters = parameters
        self.control = control

        self.is_child = is_child

    def run(self):
        try:
            self.uuid = generate_curve_uuid()
            self.set_decimation(ALL_DECIMATIONS[0])
            self.add_callbacks()
            self.parameters.psd_acquisition_running.value = True
        except Exception as e:
            self.cleanup()
            raise e

    def add_callbacks(self):
        self.parameters.acquisition_raw_data.add_callback(
            self.react_to_new_signal, call_with_first_value=False
        )

    def cleanup(self):
        self.running = False
        self.parameters.psd_acquisition_running.value = False

        self.parameters.acquisition_raw_data.remove_callback(self.react_to_new_signal)

        if not self.is_child:
            self.control.exposed_pause_acquisition()
            self.parameters.acquisition_raw_enabled.value = False
            self.parameters.acquisition_raw_filter_enabled.value = False

            self.control.exposed_write_registers()
            self.control.exposed_continue_acquisition()

    def react_to_new_signal(self, data_pickled):
        try:
            if not self.running or self.parameters.pause_acquisition.value:
                return

            data = pickle.loads(data_pickled)

            current_decimation = self.parameters.acquisition_raw_decimation.value
            logger.debug("recorded signal for decimation %s" % current_decimation)
            logger.debug("recording took %s s" % (time() - self.time_decimation_set))
            self.recorded_signals_by_decimation[current_decimation] = data
            self.recorded_psds_by_decimation[current_decimation] = residual_freq_noise(
                1 / (125e6) * (2 ** (current_decimation)),
                data[0],
                algorithm=self.parameters.psd_algorithm.value,
            )

            # this means that measurement time increases by a factor of 16 after
            # each measurement
            self.decimation_index += 4

            complete = (
                self.decimation_index
                > self.parameters.psd_acquisition_max_decimation.value
            )

            self.publish_psd_data(complete)

            if not complete:
                new_decimation = self.decimation_index
                logger.debug("set new decimation %s" % new_decimation)
                self.set_decimation(new_decimation)
            else:
                self.cleanup()

        except Exception as e:
            self.cleanup()
            raise e

    def publish_psd_data(self, complete):
        data_pickled = pickle.dumps(
            {
                "uuid": self.uuid,
                "time": time(),
                "p": self.parameters.p.value,
                "i": self.parameters.i.value,
                "d": self.parameters.d.value,
                "signals": self.recorded_signals_by_decimation,
                "psds": self.recorded_psds_by_decimation,
                "fitness": psds_to_fitness(self.recorded_psds_by_decimation),
                "complete": complete,
            }
        )
        self.parameters.psd_data_partial.value = data_pickled
        if complete:
            # we have an extra parameter for complete psd data because partial psd data
            # may change quickly. This makes it possible that someone listening to
            # partial psd data may miss the complete data set because a new partial
            # trace is being recorded.
            self.parameters.psd_data_complete.value = data_pickled

    def set_decimation(self, decimation):
        self.time_decimation_set = time()
        self.control.exposed_pause_acquisition()
        self.parameters.acquisition_raw_decimation.value = decimation
        self.parameters.acquisition_raw_enabled.value = True

        # lowpass filter for preventing alias effects
        self.parameters.acquisition_raw_filter_enabled.value = True
        self.parameters.acquisition_raw_filter_frequency.value = int(
            125e6 / (2**decimation) / 2
        )

        self.control.exposed_write_registers()
        # take care that new decimation was actually written to FPGA
        sleep(0.1)
        self.control.exposed_continue_acquisition()

    def exposed_stop(self):
        self.cleanup()


class PIDOptimization:
    def __init__(self, control, parameters):
        self.control = control
        self.parameters = parameters

        self.engine = MultiDimensionalOptimizationEngine(
            [[100, 4000], [100, 4000]], x0=[2000, 2000]
        )

    def run(self):
        try:
            self.parameters.psd_data_complete.add_callback(
                self.psd_data_received, call_with_first_value=False
            )
            self.parameters.psd_optimization_running.value = True
            self.start_single_psd_measurement()
        except Exception as e:
            self.cleanup()
            raise e

    def start_single_psd_measurement(self):
        new_params = self.engine.ask()
        self.parameters.p.value = int(new_params[0])
        self.parameters.i.value = int(new_params[1])

        self.psd_acquisition = PSDAcquisition(
            self.control, self.parameters, is_child=True
        )
        self.psd_acquisition.run()

    def cleanup(self):
        self.parameters.psd_optimization_running.value = False
        self.parameters.psd_data_complete.remove_callback(self.psd_data_received)

    def psd_data_received(self, psd_data_pickled):
        try:
            # psd data doesn't have to be stored here as a client that is interested
            # in it may listen to parameters.psd_data change events
            psd_data = pickle.loads(psd_data_pickled)

            params = (psd_data["p"], psd_data["i"])
            logger.debug("received fitness %s, %s" % (psd_data["fitness"], params))

            self.engine.tell(psd_data["fitness"], params)

            self.start_single_psd_measurement()

            done = self.engine.finished()
            if done:
                # FIXME: implement use of optimized pid parameters
                self.cleanup()

        except Exception as e:
            self.cleanup()
            raise e

    def exposed_stop(self):
        self.cleanup()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy import optimize, stats

# after the line was centered, its width will be 1/FINAL_ZOOM_FACTOR of the
# view.
FINAL_ZOOM_FACTOR = 20


def get_max_slope(signal, final_zoom_factor):
    line_width = len(signal) / final_zoom_factor
    window_width = 1.5 * line_width
    center = len(signal) / 2
    range_around_center = [
        round(center - (window_width / 2)),
        round(center + (window_width / 2)),
    ]
    crop = signal[range_around_center[0] : range_around_center[1]]

    idx = list(sorted([np.argmax(crop), np.argmin(crop)]))

    crop_of_crop = crop[idx[0] : idx[1]]

    slope, intercept, r_value, p_value, std_err = stats.linregress(
        list(range(len(crop_of_crop))), crop_of_crop
    )

    # x_diff = np.abs(idx[0] - idx[1])
    # return abs(slope * x_diff)
    return abs(slope)


def calculate_spectrum_from_iq(i, q, phase):
    return np.array(i) * np.cos(phase / 360 * 2 * np.pi) + np.array(q) * np.sin(
        phase / 360 * 2 * np.pi
    )


def optimize_phase_from_iq(i, q, final_zoom_factor):
    def iq2slope(phase):
        calculated = calculate_spectrum_from_iq(i, q, phase)
        return get_max_slope(calculated, final_zoom_factor)

    min_result = optimize.minimize_scalar(
        lambda phase: -1 * iq2slope(phase), method="Bounded", bounds=(0, 360)
    )
    assert min_result.success

    return min_result.x, abs(min_result.fun)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import math
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import cma

import logging

from linien_common.common import MHz, Vpp
from linien_server.optimization.utils import (
    FINAL_ZOOM_FACTOR,
    get_max_slope,
    optimize_phase_from_iq,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class NoOptimizationEngine:
    def __init__(self, *args):
        pass

    def finished(self):
        return True

    def ask(self):
        raise NotImplementedError()

    def tell(self, *args):
        raise NotImplementedError()


class OneDimensionalOptimizationEngine:
    def __init__(self, bounds):
        bounds = list(bounds)
        bounds.append([0.0, 1.0])
        self._multi = MultiDimensionalOptimizationEngine(bounds)
        self._last_additional_param = 0.5

    def ask(self):
        params = self._multi.ask()
        self._last_additional_param = params[1]
        return [params[0]]

    def finished(self):
        return self._multi.finished()

    def tell(self, fitness, parameters):
        parameters = list(parameters)
        parameters.append(self._last_additional_param)
        self._multi.tell(fitness, parameters)


class MultiDimensionalOptimizationEngine:
    def __init__(self, bounds, x0=None):
        self.bounds = bounds

        if x0 is not None:
            x0_converted = self.params_to_internal(x0)
        else:
            x0_converted = [0.5 for v in bounds]

        self.es = cma.CMAEvolutionStrategy(
            x0_converted,
            0.5,
            {"bounds": [[0 for v in bounds], [1 for v in bounds]]},
        )

        self._pending = []
        self._done = []
        self._results = []

    def params_to_internal(self, parameters):
        new_parameters = []
        for [min_, max_], param in zip(self.bounds, parameters):
            new_parameters.append((param - min_) / (max_ - min_))
        return new_parameters

    def internal_to_params(self, internal):
        parameters = []
        for [min_, max_], param in zip(self.bounds, internal):
            parameters.append((param * (max_ - min_)) + min_)
        return parameters

    def finished(self):
        return self.es.stop()

    def ask(self):
        if not self._pending:
            self._pending = self.es.ask()

        return self.internal_to_params(self._pending.pop())

    def tell(self, fitness, parameters):
        self._results.append(fitness)
        self._done.append(parameters)

        if not self._pending:
            self.es.tell(
                [self.params_to_internal(p) for p in self._done], self._results
            )
            self._results = []
            self._done = []


class OptimizerEngine:
    def __init__(self, control, params):
        self.control = control
        self.parameters = params

        self.init_opt_with_bounds()

        self.all_params = [
            params.modulation_frequency,
            params.modulation_amplitude,
            self.get_demod_phase_param(),
        ]
        self.params_before_start = [p.value for p in self.all_params]

        self.parameters.optimization_optimized_parameters.value = (
            self.params_before_start
        )

        self.initial_slope = None
        self.last_parameters = None
        self.last_parameters_internal = None

    def init_opt_with_bounds(self):
        params = self.parameters

        self.to_optimize = []
        self.bounds = []

        if params.optimization_mod_freq_enabled.value:
            self.to_optimize.append(params.modulation_frequency)
            freqs = list(
                sorted(
                    [
                        params.optimization_mod_freq_min.value * MHz,
                        params.optimization_mod_freq_max.value * MHz,
                    ]
                )
            )
            self.bounds.append(freqs)

        if params.optimization_mod_amp_enabled.value:
            self.to_optimize.append(params.modulation_amplitude)
            ampls = list(
                sorted(
                    [
                        params.optimization_mod_amp_min.value * Vpp,
                        params.optimization_mod_amp_max.value * Vpp,
                    ]
                )
            )
            self.bounds.append(ampls)

        self.opt = [
            NoOptimizationEngine,
            OneDimensionalOptimizationEngine,
            MultiDimensionalOptimizationEngine,
        ][len(self.bounds)]([[0, 1]] * len(self.bounds))

    def request_and_set_new_parameters(self, use_initial_parameters=False):
        self.control.exposed_pause_acquisition()

        if use_initial_parameters:
            for param, initial in zip(self.all_params, self.params_before_start):
                param.value = initial
        else:
            new_params = self.opt.ask()
            new_params_converted = [
                self.bounds[idx][0] + p * (self.bounds[idx][1] - self.bounds[idx][0])
                for idx, p in enumerate(new_params)
            ]

            for param, value in zip(self.to_optimize, new_params_converted):
                param.value = value

            self.last_parameters = list(new_params_converted)
            self.last_parameters_internal = list(new_params)

        self.control.exposed_write_registers()
        self.control.exposed_continue_acquisition()

    def finished(self):
        return self.opt.finished()

    def get_demod_phase_param(self):
        params = self.parameters
        dual_channel = params.dual_channel.value
        channel = params.optimization_channel.value

        return (
            self.parameters.demodulation_phase_a,
            self.parameters.demodulation_phase_b,
        )[0 if not dual_channel else (0, 1)[channel]]

    def tell(self, i, q):
        if self.initial_slope is None:
            self.initial_slope = get_max_slope(i, FINAL_ZOOM_FACTOR)

        optimized_phase, optimized_slope = optimize_phase_from_iq(
            i, q, FINAL_ZOOM_FACTOR
        )
        old_phase_value = self.get_demod_phase_param().value
        new_phase_value = old_phase_value - optimized_phase
        if new_phase_value > 360:
            new_phase_value -= 360
        if new_phase_value < 0:
            new_phase_value = 360 - abs(new_phase_value)

        improvement = (optimized_slope - self.initial_slope) / self.initial_slope
        params = self.parameters
        if improvement > 0 and improvement > params.optimization_improvement.value:
            params.optimization_improvement.value = improvement
            complete_parameter_set = self.params_before_start[:]
            complete_parameter_set[2] = new_phase_value

            if self.last_parameters:
                for param, value in zip(self.to_optimize, self.last_parameters):
                    idx = {
                        self.parameters.modulation_frequency: 0,
                        self.parameters.modulation_amplitude: 1,
                    }[param]
                    complete_parameter_set[idx] = value

            params.optimization_optimized_parameters.value = complete_parameter_set

        logger.debug("improvement %d" % (improvement * 100))

        fitness = math.log(1 / optimized_slope)

        if self.last_parameters_internal is not None:
            self.opt.tell(fitness, self.last_parameters_internal)

    def use_best_parameters(self):
        optimized = self.parameters.optimization_optimized_parameters.value

        for param, value in zip(self.all_params, optimized):
            param.value = value



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

"""Module providing the Optimizer interface class."""

from typing import List, NewType

Params = NewType("Params", List[float])
"""A type alias for actual optmizer params. """


class Optimizer:
    """Interface for the "Optimizer" type classes."""

    def __init__(self):
        self._generation = 0
        """Generation counter."""

        self._upper_limits = []
        """Upper parameter limit for each dimension."""

        self._lower_limits = []
        """Lower parameter limit for each dimension."""

        self._boundary_conditions = []
        """Complex boundary conditions list"""

    def request_parameter_set(self):
        """Ask the engine for a new param set to try out."""

        raise NotImplementedError("request_parameter_set is not implemented.")
        return []

    def insert_fitness_value(self, fitness, set):
        """Tell the engine about our experiment result."""
        raise NotImplementedError("insert_fitness_value not implemented")

    def request_results(self):
        """Return a list of promising results."""
        raise NotImplementedError("request_results not implemented")
        return []

    @property
    def generation(self):
        return self._generation

    def _truncate_parameters(self, params):
        """
        Make sure the passed list of parameters is within [min, max] if available,
        respectively.
        """

        if self._lower_limits != []:
            params = [
                max(low_lim, par) for low_lim, par in zip(self._lower_limits, params)
            ]

        if self._upper_limits != []:
            params = [
                min(low_lim, par) for low_lim, par in zip(self._upper_limits, params)
            ]

        if self._boundary_conditions != []:
            for condition in self._boundary_conditions:
                params = condition.check_and_correct(params)

        return params


class Individual:
    """A parameter set of an optimization algorithm."""

    def __init__(self, param_set, fitness):
        self.param_set = param_set
        self.fitness = fitness
        self.generation = 0

    def __lt__(self, other):
        return self.fitness < other.fitness

    def __gt__(self, other):
        return self.fitness > other.fitness

    def __copy__(self):
        ind = Individual(self.param_set.copy(), self.fitness)
        ind.generation = self.generation
        return ind



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
# Copyright 2023 Christian Freier <christian.freier@nomadatomics.com>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from time import time
from typing import Any, Callable, Dict, Iterator, List, Tuple

import linien_server
from linien_common.common import AutolockMode, MHz, PSDAlgorithm, Vpp
from linien_common.config import USER_DATA_PATH

PARAMETER_STORE_FILENAME = "linien_parameters.json"

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Parameter:
    """Represents a single parameter and is used by `Parameters`."""

    def __init__(
        self,
        min_=None,
        max_=None,
        start=None,
        wrap=False,
        sync=True,
        collapsed_sync=True,
        restorable=False,
        loggable=False,
        log=False,
    ):
        self.min = min_
        self.max = max_
        self.wrap = wrap
        self._value = start
        self._start = start
        self._callbacks = set()
        self.can_be_cached = sync
        self._collapsed_sync = collapsed_sync
        self.restorable = restorable
        self.loggable = loggable
        self.log = log

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        # check bounds
        if self.min is not None and value < self.min:
            value = self.min if not self.wrap else self.max
        if self.max is not None and value > self.max:
            value = self.max if not self.wrap else self.min

        self._value = value

        # We copy it because a listener could remove a listener --> this would cause an
        # error in this loop.
        for callback in self._callbacks.copy():
            callback(value)

    def reset(self):
        self.value = self._start

    def add_callback(
        self,
        function: Callable[[Any], None],
        call_with_first_value: bool = True,
    ) -> None:
        self._callbacks.add(function)

        if call_with_first_value:
            if self._value is not None:
                function(self._value)

    def remove_callback(self, function: Callable[[Any], None]) -> None:
        if function in self._callbacks:
            self._callbacks.remove(function)


class Parameters:
    """
    This class defines the parameters of the Linien server. They represent the public
    interface and can be used to control the behavior of the server.

    Access on the server takes place like this:

        # retrieve a parameter
        foo(parameters.modulation_amplitude.value)

        # set a parameter
        parameters.modulation_amplitude.value = 0.5 * Vpp

        # if a parameter influences the behavior of the FPGA, it has to be
        # written to the FPGA as well (`control` is an instance of
        # `RedPitayaControlService`):
        control.exposed_write_registers()

    On the client side, access happens through `RemoteParameters` which
    transparently mimics the behavior of this class. Have a look at the comments
    below for a description of each parameter.
    """

    def __init__(self):
        # Dict[str, List[Tuple[str, Any]]]
        self._changed_parameters_queue = {}
        # Dict[Tuple[Parameter, Callable[[Any], None]]]
        self._remote_listener_callbacks = {}

        self.to_plot = Parameter(sync=False)
        """
        The `to_plot` parameter is a pickled dictionary that contains signalsvthat may
        be plotted. Depending on the locking state, it may contain thesevsignals:
        Unlocked state:
          - `error_signal_1` and `error_signal_1_quadrature`:
              IQ-demodulated and low-pass-filtered error signals from ANALOG IN 0
          - `error_signal_2` and `error_signal_2_quadrature`:
              IQ-demodulated and low-pass-filtered error signals from ANALOG IN 1. These
              signals are only available if in dual-channel spectroscopy mode is
              enabled. Otherwise the un-demodulated monitor_signal` is shown.
          - `monitor_signal`:
              Signal recorded from ANALOG IN 1 without demodulation. Only available if
              dual channel mode is not enabled.
        Locked state:
          - `error_signal`:
              Error signal that is fed into the PID controller.
          - `control_signal`:
              Output of the PID controller.
          - `slow`:
              Output of slow additional integrator on slow analog output. Note that this
              value is not an array but a single point.
        """

        self.signal_stats = Parameter(sync=False, loggable=True)
        """
        A dictionary that contains mean, standard deviation, minimum value and maximum
        value for all the signals contained in `to_plot`. Exemplary dictionary keys are
        `error_signal_mean`, `control_signal_std`, `monitor_signal_min` or
        `error_signal_2_max`.
        """

        # ------------------- GENERAL PARAMETERS ---------------------------------------

        self.mod_channel = Parameter(start=0, min_=0, max_=1, restorable=True)
        """
        Configures the output of the modulation frequency. A value of 0 means FAST OUT 1
        and a value of 1 corresponds to FAST OUT 2
        """

        self.sweep_channel = Parameter(start=1, min_=0, max_=2, restorable=True)
        """
        Configures the output of the scan sweep:
            0 --> FAST OUT 1
            1 --> FAST OUT 2
            2 --> ANALOG OUT 0 (slow channel)
        """

        self.control_channel = Parameter(start=1, min_=0, max_=1, restorable=True)
        """
        Configures the output of the lock signal. A value of 0 means FAST OUT 1 and a
        value of 1 corresponds to FAST OUT 2
        """

        self.slow_control_channel = Parameter(start=2, min_=0, max_=2, restorable=True)
        """
        Configures the output of the slow PID control:
            0 --> FAST OUT 1
            1 --> FAST OUT 2
            2 --> ANALOG OUT 0 (slow channel)
        """

        self.gpio_p_out = Parameter(start=0, min_=0, max_=0b11111111)
        """
        set the output of GPIO pins. Each bit corresponds to one pin, i.e.
        `parameters.gpio_p_out.value = 0b11110000` turns on the first 4 pins and turns
        off the other ones.
        """
        self.gpio_n_out = Parameter(start=0, min_=0, max_=0b11111111)
        """
        set the output of GPIO pins. Each bit corresponds to one pin, i.e.
        `parameters.gpio_p_out.value = 0b11110000` turns on the first 4 pins and turns
        off the other ones.
        """

        # ANALOG_OUT0 is used for slow PID --> it can't be controlled manually
        self.analog_out_1 = Parameter(
            start=0, min_=0, max_=(2**15) - 1, restorable=True
        )
        """
        parameters for setting ANALOG_OUT 1 voltage.
        Usage: `parameters.analog_out_1.value = 1.2 * ANALOG_OUT_V`
        Minimum value is 0 and maximum 1.8 * ANALOG_OUT_V
        """
        self.analog_out_2 = Parameter(
            start=0, min_=0, max_=(2**15) - 1, restorable=True
        )
        """
        parameters for setting ANALOG_OUT 1 voltage.
        Usage: `parameters.analog_out_2.value = 1.2 * ANALOG_OUT_V`
        Minimum value is 0 and maximum 1.8 * ANALOG_OUT_V
        """
        self.analog_out_3 = Parameter(
            start=0, min_=0, max_=(2**15) - 1, restorable=True
        )
        """
        parameters for setting ANALOG_OUT 1 voltage.
        Usage: `parameters.analog_out_3.value = 1.2 * ANALOG_OUT_V`
        Minimum value is 0 and maximum 1.8 * ANALOG_OUT_V
        """

        self.lock = Parameter(start=False, loggable=True)
        """If `True`, this parameter turns off the sweep and starts the PID"""

        self.polarity_fast_out1 = Parameter(start=False, restorable=True)
        """
        Defines whether tuning the voltage up correspond to tuning the laser frequency
        up or down. Setting these values correctly is only required when using both, a
        fast out and a the slow output for PID
        """

        self.polarity_fast_out2 = Parameter(start=False, restorable=True)
        """
        Defines whether tuning the voltage up correspond to tuning the laser frequency
        up or down. Setting these values correctly is only required when using both, a
        fast out and a the slow output for PID
        """

        self.polarity_analog_out0 = Parameter(start=False, restorable=True)
        """
        Defines whether tuning the voltage up correspond to tuning the laser frequency
        up or down. Setting these values correctly is only required when using both, a
        fast out and a the slow output for PID
        """

        self.control_signal_history_length = Parameter(start=600)
        """Record of control signal should be kept for how long?"""

        self.control_signal_history = Parameter(
            start={"times": [], "values": []}, sync=False
        )

        self.monitor_signal_history = Parameter(
            start={"times": [], "values": []}, sync=False
        )

        self.pause_acquisition = Parameter(start=False)
        """
        If this boolean is `True`, no new spectroscopy data is sent to the clients. This
        parameter is used when writing data to FPGA that would result in cropped or
        distorted signals being displayed.
        """

        self.fetch_additional_signals = Parameter(start=True)
        """
        This parameter is not exposed to GUI. It is used by the autolock or normal lock
        to fetch less data if they are not needed.
        """

        self.ping = Parameter(start=0)
        """
        This is just a counter that is automatically increased every second. Its purpose
        is to allow for periodic tasks on the server: just register a callback with
         `add_callback` for this parameter.
        """

        # ------------------- SWEEP PARAMETERS -----------------------------------------

        self.sweep_amplitude = Parameter(min_=0.001, max_=1, start=1, loggable=True)
        """
        Amplitude of the sweep in units of 0.5 * Vpp of the output (2 V for fast outputs
        (range +/- 1 V) and 0.9 V for slow outputs (range 0 V to 1.8 V). That means an
        amplitude of 1.0 corresponds to the full sweep range in both cases.
        """

        self.sweep_center = Parameter(min_=-1, max_=1, start=0, loggable=True)
        """
        The center position of the sweep. If a fast output is used for the sweep this is
        the sweep center position in volts. If the slow output is used the interval
        [-1, +1] of this parameter is mapped to the interval [0V, +1.8V].
        """

        self.sweep_speed = Parameter(
            min_=0, max_=32, start=8, restorable=True, loggable=True
        )
        """
        The sweep speed in internal units. The actual speed is given by
        f_real = 3.8 kHz / (2 ** sweep_speed)
        Allowed values are [0, ..., 16]
        """

        self.sweep_pause = Parameter(start=False, loggable=True)
        """If set to `True`, this parameter pauses the sweep at the `sweep_center`."""

        # ------------------- MODULATION PARAMETERS ------------------------------------

        self.modulation_amplitude = Parameter(
            min_=0, max_=(1 << 14) - 1, start=1 * Vpp, restorable=True, loggable=True
        )
        """
        The amplitude of the modulation in internal units. Use Vpp for conversion to
        volts peak-peak, e.g: `parameters.modulation_amplitude.value = 0.5 * Vpp`
        Values between 0 and 2 * Vpp are allowed.
        """

        self.modulation_frequency = Parameter(
            min_=0, max_=0xFFFFFFFF, start=15 * MHz, restorable=True, loggable=True
        )
        """
        Frequency of the modulation in internal units. Use MHz for conversion to
        human-readable frequency, e.g:
            `parameters.modulation_frequency.value = 6.6 * MHz`
        By design, values up to 128 * MHz = 0xffffffff are allowed although in practice
        values of more than 50 MHz don't make sense due to the limited sampling rate of
        the DAC.
        """

        # ------------------- DEMODULATION AND FILTER PARAMETERS -----------------------
        self.pid_only_mode = Parameter(start=False, restorable=True)
        """
        PID-only mode allows to bypass demodulation, IIR filtering and offset.
        """

        self.dual_channel = Parameter(start=False, restorable=True)
        """
        Linien allows for two simultaneous demodulation channels. By default, only one
        is enabled. This is controlled by `dual_channel`.
        """

        self.channel_mixing = Parameter(start=0, restorable=True, loggable=True)
        """
        If in dual channel mode, what is the mixing ratio between them? A value of 0
        corresponds to equal ratio
                   -128             only channel A being active
                   128              only channel B being active
        Integer values [-128, ..., 128] are allowed.
        """

        # The following parameters exist twice, i.e. once per channel
        self.demodulation_phase_a = Parameter(
            min_=0, max_=360, start=0x0, wrap=True, restorable=True, loggable=True
        )
        """The demodulation phase for channel A in degree (0-360)"""

        self.demodulation_phase_b = Parameter(
            min_=0, max_=360, start=0x0, wrap=True, restorable=True, loggable=True
        )
        """The demodulation phase for channel B in degree (0-360)"""

        self.demodulation_multiplier_a = Parameter(
            min_=0, max_=15, start=1, restorable=True, loggable=True
        )
        """
        This parameter allows for multi-f (e.g. 3f or 5f) demodulation. Default value is
        1, indicating that 1f demodulation is used.
        """

        self.demodulation_multiplier_b = Parameter(
            min_=0, max_=15, start=1, restorable=True, loggable=True
        )
        """
        This parameter allows for multi-f (e.g. 3f or 5f) demodulation. Default value is
        1, indicating that 1f demodulation is used.
        """

        self.offset_a = Parameter(
            min_=-8191, max_=8191, start=0, restorable=True, loggable=True
        )
        """
        The vertical offset for channel A. A value of -8191 shifts the data down by 1V,
        a value of +8191 moves it up.
            """

        self.offset_b = Parameter(
            min_=-8191, max_=8191, start=0, restorable=True, loggable=True
        )
        """
        The vertical offset for channel B. A value of -8191 shifts the data down by 1V,
        a value of +8191 moves it up.
            """

        self.invert_a = Parameter(start=False, restorable=True)
        """A boolean indicating whether the channel A data should be inverted."""

        self.invert_b = Parameter(start=False, restorable=True)
        """A boolean indicating whether the channel B data should be inverted."""

        # -------   FILTER PARAMETERS   ------------------------------------------------
        self.filter_automatic_a = Parameter(start=True, restorable=True)
        """
        After demodulation of the signal, Linien may apply up to two IIR filters.
        `filter_automatic` is a boolean indicating whether Linien should automatically
        determine suitable filter for a given modulation frequency or whether the user
        may configure the filters himself. If automatic mode is enabled, two low pass
        filters are installed with a frequency of half the modulation frequency.
        """

        self.filter_automatic_b = Parameter(start=True, restorable=True)
        """
        After demodulation of the signal, Linien may apply up to two IIR filters.
        `filter_automatic` is a boolean indicating whether Linien should automatically
        determine suitable filter for a given modulation frequency or whether the user
        may configure the filters himself. If automatic mode is enabled, two low pass
        filters are installed with a frequency of half the modulation frequency.
        """

        self.filter_1_enabled_a = Parameter(start=False, restorable=True)
        """
        Should this filter be enabled? Note that disabling a filter does not bypass it
        as this would change the propagation time of the signal through the FPGA which
        is unfavorable as it leads to a mismatch of the demodulation phase. Instead, a
        filter with unity gain is installed.
        """

        self.filter_2_enabled_a = Parameter(start=False, restorable=True)
        """
        Should this filter be enabled? Note that disabling a filter does not bypass it
        as this would change the propagation time of the signal through the FPGA which
        is unfavorable as it leads to a mismatch of the demodulation phase. Instead, a
        filter with unity gain is installed.
        """

        self.filter_1_enabled_b = Parameter(start=False, restorable=True)
        """
        Should this filter be enabled? Note that disabling a filter does not bypass it
        as this would change the propagation time of the signal through the FPGA which
        is unfavorable as it leads to a mismatch of the demodulation phase. Instead, a
        filter with unity gain is installed.
        """

        self.filter_2_enabled_b = Parameter(start=False, restorable=True)
        """
        Should this filter be enabled? Note that disabling a filter does not bypass it
        as this would change the propagation time of the signal through the FPGA which
        is unfavorable as it leads to a mismatch of the demodulation phase. Instead, a
        filter with unity gain is installed.
        """

        self.filter_1_frequency_a = Parameter(start=10000, restorable=True)
        """The filter frequency in units of Hz"""

        self.filter_2_frequency_a = Parameter(start=10000, restorable=True)
        """The filter frequency in units of Hz"""

        self.filter_1_frequency_b = Parameter(start=10000, restorable=True)
        """The filter frequency in units of Hz"""

        self.filter_2_frequency_b = Parameter(start=10000, restorable=True)
        """The filter frequency in units of Hz"""

        self.filter_1_type_a = Parameter(start=0, restorable=True)
        """
        Either `LOW_PASS` or `HIGH_PASS` of linien_common.common.FilterType` enum class.
        """

        self.filter_2_type_a = Parameter(start=0, restorable=True)
        """
        Either `LOW_PASS` or `HIGH_PASS` of linien_common.common.FilterType` enum class.
        """

        self.filter_1_type_b = Parameter(start=0, restorable=True)
        """
        Either `LOW_PASS` or `HIGH_PASS` of linien_common.common.FilterType` enum class.
        """

        self.filter_2_type_b = Parameter(start=0, restorable=True)
        """
        Either `LOW_PASS` or `HIGH_PASS` of linien_common.common.FilterType` enum class.
        """

        # ------------------- LOCK AND PID PARAMETERS ----------------------------------

        self.combined_offset = Parameter(min_=-8191, max_=8191, start=0)
        """
        After combining channels A and B and before passing the result to the PID,
        `combined_offset` is added. It uses the same units as the channel offsets, i.e.
        a value of -8191 shifts the data down by 1V, a value of +8191 moves it up.
        """

        self.p = Parameter(start=50, max_=8191, restorable=True, loggable=True)
        """
        Proportional part of PID parameters. Range is [0, 8191]. In order to change sign
        of PID parameters, use `target_slope_rising`
        """

        self.i = Parameter(start=5, max_=8191, restorable=True, loggable=True)
        """
        Integral part of PID parameters. Range is [0, 8191]. In order to change sign of
        PID parameters, use `target_slope_rising`
        """

        self.d = Parameter(start=0, max_=8191, restorable=True, loggable=True)
        """
        Derivate part of PID parameters. Range is [0, 8191]. In order to change sign of
        PID parameters, use `target_slope_rising`
        """

        self.target_slope_rising = Parameter(start=True)
        """A boolean that inverts the sign of the PID parameters"""

        self.pid_on_slow_enabled = Parameter(start=False, restorable=True)
        """Whether the PID on ANALOG_OUT 0 is enabled."""

        self.pid_on_slow_strength = Parameter(start=0, restorable=True)
        """
        Strength of the slow PID. This strength corresponds to the strength of the
        integrator. Maximum value is 8191.
        """

        self.check_lock = Parameter(start=True, restorable=True)
        self.watch_lock = Parameter(start=True, restorable=True)
        self.watch_lock_threshold = Parameter(start=0.01, restorable=True)

        # ------------------- AUTOLOCK PARAMETERS --------------------------------------
        # These parameters are used internally by the optimization algorithm and usually
        # should not be manipulated
        self.task = Parameter(start=None, sync=False)
        self.automatic_mode = Parameter(start=True)
        self.autolock_target_position = Parameter(start=0)
        self.autolock_mode_preference = Parameter(
            start=AutolockMode.AUTO_DETECT, restorable=True
        )
        self.autolock_mode = Parameter(start=AutolockMode.SIMPLE)
        self.autolock_time_scale = Parameter(start=0)
        self.autolock_instructions = Parameter(start=[], sync=False)
        self.autolock_final_wait_time = Parameter(start=0)
        self.autolock_selection = Parameter(start=False)
        self.autolock_running = Parameter(start=False)
        self.autolock_preparing = Parameter(start=False)
        self.autolock_percentage = Parameter(start=0, min_=0, max_=100)
        self.autolock_watching = Parameter(start=False)
        self.autolock_failed = Parameter(start=False)
        self.autolock_locked = Parameter(start=False)
        self.autolock_retrying = Parameter(start=False)
        self.autolock_determine_offset = Parameter(start=True, restorable=True)
        self.autolock_initial_sweep_amplitude = Parameter(start=1)

        # ------------------- OPTIMIZATION PARAMETERS ----------------------------------
        # These parameters are used internally by the optimization algorithm and usually
        # should not be manipulated
        self.optimization_selection = Parameter(start=False)
        self.optimization_running = Parameter(start=False)
        self.optimization_approaching = Parameter(start=False)
        self.optimization_improvement = Parameter(start=0)
        self.optimization_mod_freq_enabled = Parameter(start=1)
        self.optimization_mod_freq_min = Parameter(start=0.0)
        self.optimization_mod_freq_max = Parameter(start=10.0)
        self.optimization_mod_amp_enabled = Parameter(start=1)
        self.optimization_mod_amp_min = Parameter(start=0.0)
        self.optimization_mod_amp_max = Parameter(start=2.0)
        self.optimization_optimized_parameters = Parameter(start=(0, 0, 0))
        self.optimization_channel = Parameter(start=0)
        self.optimization_failed = Parameter(start=False)

        # ------------------- PID OPTIMIZATION PARAMETERS ------------------------------
        # These parameters are used internally by the optimization algorithm and usually
        # should not be manipulated
        self.acquisition_raw_enabled = Parameter(start=False)
        self.acquisition_raw_decimation = Parameter(start=1)
        self.acquisition_raw_data = Parameter()
        self.acquisition_raw_filter_enabled = Parameter(start=False)
        """
        Raw acquisition has an additional iir filter that can be used as low pass for
        preventing alias effects
        """

        self.acquisition_raw_filter_frequency = Parameter(start=0)
        """The filter frequency in units of Hz"""

        self.psd_data_partial = Parameter(start=None)
        self.psd_data_complete = Parameter(start=None)
        self.psd_algorithm = Parameter(start=PSDAlgorithm.LPSD)
        self.psd_acquisition_running = Parameter(start=False)
        self.psd_optimization_running = Parameter(start=False)
        self.psd_acquisition_max_decimation = Parameter(
            start=18, min_=1, max_=32, restorable=True
        )

    def __iter__(self) -> Iterator[Tuple[str, Parameter]]:
        for name, param in self.__dict__.items():
            if isinstance(param, Parameter):
                yield name, param

    def init_parameter_sync(
        self, uuid: str
    ) -> Iterator[Tuple[str, Any, bool, bool, bool, bool]]:
        """
        To be called by a remote client: Yields all parameters as well as their values
        and if the parameters are suited to be cached registers a listener that pushes
        changes of these parameters to the client.
        """
        for name, param in self:
            yield (
                name,
                param.value,
                param.can_be_cached,
                param.restorable,
                param.loggable,
                param.log,
            )
            if param.can_be_cached:
                self.register_remote_listener(uuid, name)

    def register_remote_listener(self, uuid: str, param_name: str) -> None:
        self._changed_parameters_queue.setdefault(uuid, [])
        self._remote_listener_callbacks.setdefault(uuid, [])

        def append_changed_values_to_queue(value: Any) -> None:
            """Appends changed values to the queue of a specific client."""
            if uuid in self._changed_parameters_queue:
                self._changed_parameters_queue[uuid].append((param_name, value))

        param = getattr(self, param_name)
        param.add_callback(append_changed_values_to_queue)

        self._remote_listener_callbacks[uuid].append(
            (param, append_changed_values_to_queue)
        )

    def unregister_remote_listeners(self, uuid: str):
        for param, callback in self._remote_listener_callbacks[uuid]:
            param.remove_callback(callback)

        del self._changed_parameters_queue[uuid]
        del self._remote_listener_callbacks[uuid]

    def get_changed_parameters_queue(
        self, uuid: str
    ) -> Dict[str, List[Tuple[str, Any]]]:
        """Get the queue of parameter changes for a specific client."""
        queue = self._changed_parameters_queue.get(uuid, [])
        self._changed_parameters_queue[uuid] = []

        # filter out multiple values for collapsible parameters
        already_has_value = []
        for idx in reversed(range(len(queue))):
            param_name, value = queue[idx]
            if getattr(self, param_name)._collapsed_sync:
                if param_name in already_has_value:
                    del queue[idx]
                else:
                    already_has_value.append(param_name)
        return queue


def restore_parameters(parameters: Parameters) -> Parameters:
    """When the server starts, this method restores previously saved parameters."""
    filename = str(USER_DATA_PATH / PARAMETER_STORE_FILENAME)
    try:
        with open(filename, "r") as f:
            data = json.load(f)
    except FileNotFoundError:
        return parameters

    for name, attributes in data["parameters"].items():
        try:
            getattr(parameters, name).value = attributes["value"]
            getattr(parameters, name).log = attributes["log"]
        except AttributeError:  # ignore parameters that don't exist (anymore)
            continue
    logger.info("Restored parameters from %s" % filename)
    return parameters


def save_parameters(parameters: Parameters) -> None:
    """Gather all parameters and store them on disk."""

    parameters_dict = {}
    for name, param in parameters:
        if param.restorable:
            parameters_dict[name] = {"value": param.value, "log": param.log}

    filename = str(USER_DATA_PATH / PARAMETER_STORE_FILENAME)
    with open(filename, "w") as f:
        json.dump(
            {
                "version": linien_server.__version__,
                "time": time(),
                "parameters": parameters_dict,
            },
            f,
            indent=2,
        )
    logger.info("Saved parameters to %s" % filename)



# Copyright 2018-2023 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2023 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pickle
import shutil
import subprocess
from pathlib import Path
from random import random
from threading import Event, Thread
from time import sleep
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
from linien_common.common import DECIMATION, MAX_N_POINTS, N_POINTS
from linien_common.config import ACQUISITION_PORT
from linien_server.csr import PythonCSR
from pyrp3.board import RedPitaya  # type: ignore
from pyrp3.instrument import TriggerSource  # type: ignore
from rpyc import Service
from rpyc.utils.server import ThreadedServer

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class AcquisitionService(Service):
    def __init__(self):
        super(AcquisitionService, self).__init__()
        stop_nginx()
        flash_fpga()

        self.red_pitaya = RedPitaya()
        self.csr = PythonCSR(self.red_pitaya)
        self.csr_queue = []
        self.csr_iir_queue = []

        self.data = pickle.dumps(None)
        self.data_was_raw = False
        self.data_hash = None
        self.data_uuid = None

        self.locked = False
        self.exposed_set_sweep_speed(9)
        # when self.locked is set to True, this doesn't mean that the lock is really on.
        # It just means that the lock is requested and that the gateware waits until the
        # sweep is at the correct position for the lock. Therefore, when self.locked is
        # set, the acquisition process waits for confirmation from the gateware that the
        # lock is actually running.
        self.confirmed_that_in_lock = False

        self.fetch_additional_signals = True
        self.raw_acquisition_enabled = False
        self.raw_acquisition_decimation = 0

        self.dual_channel = False

        self.stop_event = Event()
        self.pause_event = Event()
        self.skip_next_data_event = Event()

        self.thread = Thread(
            target=self._acquisition_loop,
            args=(
                self.stop_event,
                self.pause_event,
                self.skip_next_data_event,
            ),
            daemon=True,
        )
        self.thread.start()

    def _acquisition_loop(
        self, stop_event: Event, pause_event: Event, skip_next_data_event: Event
    ) -> None:
        while not stop_event.is_set():
            while self.csr_queue:
                key, value = self.csr_queue.pop(0)
                self.csr.set(key, value)

            while self.csr_iir_queue:
                args = self.csr_iir_queue.pop(0)
                self.csr.set_iir(*args)

            if self.locked and not self.confirmed_that_in_lock:
                self.confirmed_that_in_lock = self.csr.get(
                    "logic_autolock_lock_running"
                )
                if not self.confirmed_that_in_lock:
                    sleep(0.05)
                    continue

            if pause_event.is_set():
                sleep(0.05)
                continue

            # check that scope is triggered; copied from
            # https://github.com/RedPitaya/RedPitaya/blob/14cca62dd58f29826ee89f4b28901602f5cdb1d8/api/src/oscilloscope.c#L115  # noqa: E501
            if not (self.red_pitaya.scope.read(0x1 << 2) & 0x4) <= 0:
                sleep(0.05)
                continue

            if self.raw_acquisition_enabled:
                data_raw = self.read_data_raw(
                    0x10000, self.red_pitaya.scope.write_pointer_trigger, MAX_N_POINTS
                )
                is_raw = True
            else:
                data = self.read_data()
                is_raw = False

            if pause_event.is_set():
                # it may seem strange that we check this here a second time. Reason:
                # `read_data` takes some time and if in the mean time acquisition
                # was paused, we do not want to send the data
                continue

            if skip_next_data_event.is_set():
                skip_next_data_event.clear()
            else:
                if self.raw_acquisition_enabled:
                    self.data = pickle.dumps(data_raw)
                else:
                    self.data = pickle.dumps(data)
                self.data_was_raw = is_raw
                self.data_hash = random()

            self.program_acquisition_and_rearm()

    def read_data(self) -> Dict[str, np.ndarray]:
        signals = []

        channel_offsets = [0x10000]
        if self.fetch_additional_signals or self.locked:
            channel_offsets.append(0x20000)

        for channel_offset in channel_offsets:
            channel_data = self.read_data_raw(
                channel_offset,
                self.red_pitaya.scope.write_pointer_trigger,
                N_POINTS,
            )

            for sub_channel_idx in (0, 1):
                signals.append(channel_data[sub_channel_idx])

        signals_named = {}

        if not self.locked:
            signals_named["error_signal_1"] = signals[0]

            if self.fetch_additional_signals and len(signals) >= 3:
                signals_named["error_signal_1_quadrature"] = signals[2]

            if self.dual_channel:
                signals_named["error_signal_2"] = signals[1]
                if self.fetch_additional_signals and len(signals) >= 3:
                    signals_named["error_signal_2_quadrature"] = signals[3]
            else:
                signals_named["monitor_signal"] = signals[1]

        else:
            signals_named["error_signal"] = signals[0]
            signals_named["control_signal"] = signals[1]

            if not self.dual_channel and len(signals) >= 3:
                signals_named["monitor_signal"] = signals[2]

        slow_out = self.csr.get("logic_slow_value")
        slow_out = slow_out if slow_out <= 8191 else slow_out - 16384
        signals_named["slow_control_signal"] = slow_out

        return signals_named

    def read_data_raw(
        self, offset: int, addr: int, data_length: int
    ) -> Tuple[Any, ...]:
        max_data_length = 16383
        if data_length + addr > max_data_length:
            to_read_later = data_length + addr - max_data_length
            data_length -= to_read_later
        else:
            to_read_later = 0

        # raw data contains two signals:
        #   2'h0,adc_b_rd,2'h0,adc_a_rd
        #   i.e.: 2 zero bits, channel b (14 bit), 2 zero bits, channel a (14 bit)
        # .copy() is required, because np.frombuffer returns a readonly array
        raw_data = self.red_pitaya.scope.reads(offset + (4 * addr), data_length).copy()

        # raw_data is an array of 32-bit ints. We cast it to 16 bit --> each original
        # int is split into two ints
        raw_data.dtype = np.int16

        # sign bit is at position 14, but we have 16 bit ints
        raw_data[raw_data >= 2**13] -= 2**14

        # order is such that we have first the signal a then signal b
        signals = tuple(raw_data[signal_idx::2] for signal_idx in (0, 1))

        if to_read_later > 0:
            additional_raw_data = self.read_data_raw(offset, 0, to_read_later)
            signals = tuple(
                np.append(signals[signal_idx], additional_raw_data[signal_idx])
                for signal_idx in (0, 1)
            )

        return signals

    def program_acquisition_and_rearm(self, trigger_delay=16384):
        """Program the acquisition settings and rearm acquisition."""
        if not self.locked:
            target_decimation = 2 ** (self.sweep_speed + int(np.log2(DECIMATION)))

            self.red_pitaya.scope.data_decimation = target_decimation
            self.red_pitaya.scope.trigger_delay = int(trigger_delay / DECIMATION) - 1

        elif self.raw_acquisition_enabled:
            self.red_pitaya.scope.data_decimation = 2**self.raw_acquisition_decimation
            self.red_pitaya.scope.trigger_delay = trigger_delay

        else:
            self.red_pitaya.scope.data_decimation = 1
            self.red_pitaya.scope.trigger_delay = int(trigger_delay / DECIMATION) - 1

        self.red_pitaya.scope.rearm(trigger_source=TriggerSource.ext_posedge)

    def exposed_return_data(
        self, last_hash: Optional[float]
    ) -> Tuple[
        bool,
        Union[float, None],
        Union[bool, None],
        Union[bytes, None],
        Union[float, None],
    ]:
        no_data_available = self.data_hash is None
        data_not_changed = self.data_hash == last_hash
        if data_not_changed or no_data_available or self.pause_event.is_set():
            return False, None, None, None, None
        else:
            return True, self.data_hash, self.data_was_raw, self.data, self.data_uuid

    def exposed_set_sweep_speed(self, speed):
        self.sweep_speed = speed
        # if a slow acqisition is currently running and we change the sweep speed we
        # don't want to wait until it finishes
        self.program_acquisition_and_rearm()

    def exposed_set_lock_status(self, locked: bool) -> None:
        self.locked = locked
        self.confirmed_that_in_lock = False

    def exposed_set_fetch_additional_signals(self, fetch: bool) -> None:
        self.fetch_additional_signals = fetch

    def exposed_set_raw_acquisition(self, enabled: bool, decimation: int) -> None:
        self.raw_acquisition_enabled = enabled
        self.raw_acquisition_decimation = decimation

    def exposed_set_dual_channel(self, dual_channel):
        self.dual_channel = dual_channel

    def exposed_set_csr(self, key: str, value: int) -> None:
        self.csr_queue.append((key, value))

    def exposed_set_iir_csr(self, *args):
        self.csr_iir_queue.append(args)

    def exposed_stop_acquisition(self) -> None:
        self.stop_event.set()
        self.thread.join()
        start_nginx()

    def exposed_pause_acquisition(self):
        self.pause_event.set()
        self.data_hash = None
        self.data = None

    def exposed_continue_acquisition(self, uuid: Optional[float]) -> None:
        self.program_acquisition_and_rearm()
        sleep(0.01)
        # resetting data here is not strictly required but we want to be on the safe
        # side
        self.data_hash = None
        self.data = None
        self.pause_event.clear()
        self.data_uuid = uuid
        # if we are sweeping, we have to skip one data set because an incomplete sweep
        # may have been recorded. When locked, this does not matter
        if self.confirmed_that_in_lock:
            self.skip_next_data_event.clear()
        else:
            self.skip_next_data_event.set()


def flash_fpga():
    filepath = Path(__file__).resolve().parent / "gateware.bin"

    # On redpitaya os < 2, flashing fpga works by copying the bit file /dev/xdevcfg. On
    # recent versions, there is a dedicated command for this
    # cf. https://forum.redpitaya.com/viewtopic.php?p=33494&sid=5132bf6e33709b1a7daa948f8e8dcdb1#p33494  # noqa: E501
    fpga_dev_file = Path("/dev/xdevcfg")

    if fpga_dev_file.exists():
        logger.info("Copying gateware to %s" % fpga_dev_file)
        shutil.copy(str(filepath), str(fpga_dev_file))
    else:
        logger.info("Using fpgautil to deploy gateware.")
        subprocess.Popen(["/opt/redpitaya/bin/fpgautil", "-b", str(filepath)]).wait()


def start_nginx():
    subprocess.Popen(["systemctl", "start", "redpitaya_nginx.service"])


def stop_nginx():
    subprocess.Popen(["systemctl", "stop", "redpitaya_nginx.service"]).wait()
    subprocess.Popen(["systemctl", "stop", "redpitaya_scpi.service"]).wait()


if __name__ == "__main__":
    threaded_server = ThreadedServer(AcquisitionService(), port=ACQUISITION_PORT)
    logger.info("Starting AcquisitionService on port %s" % ACQUISITION_PORT)
    threaded_server.start()
